//
// Session.cpp
//
#include "Session.hpp"
#include "Utils.hpp"
#include <string>

namespace network {

    Session::Session(boost::asio::io_service& io_service)
    : io_service_(io_service),
      socket_(io_service),
      encryption_(false),
      online_(true),
      login_(false),
      account_revision_(0),
      read_start_time_(time(nullptr)),
      write_start_time_(time(nullptr)),
      read_byte_sum_(0),
      write_byte_sum_(0),
      serialized_byte_sum_(0),
      compressed_byte_sum_(0),
      id_(0)
    {
    }

    Session::~Session()
    {
        Close();
    }

    void Session::Close()
    {
        socket_.close();
    }

    void Session::Send(const Command& command)
    {
        auto msg = Serialize(command);
        write_byte_sum_ += msg.size();
        UpdateWriteByteAverage();

        io_service_.post(boost::bind(&Session::DoWrite, this, msg));
    }

    void Session::SyncSend(const Command& command)
    {
        auto msg = Serialize(command);
        write_byte_sum_ += msg.size();
        UpdateWriteByteAverage();

        try {
            boost::asio::write(
                    socket_, boost::asio::buffer(msg.data(), msg.size()),
                boost::asio::transfer_all());
        } catch (std::exception& e) {
            std::cout << e.what() << std::endl;
        }
    }

    void Session::LogIn(UserID id)
    {
        id_ = id;
        login_ = true;
    }

    void Session::LogOut()
    {
        login_ = false;
    }

    double Session::GetReadByteAverage() const
    {
        return 1.0f * read_byte_sum_ / (time(nullptr) - read_start_time_);
    }

    double Session::GetWriteByteAverage() const
    {
        return 1.0f * write_byte_sum_ / (time(nullptr) - write_start_time_);
    }

    void Session::UpdateReadByteAverage()
    {
        unsigned long elapsed_time = time(nullptr) - read_start_time_;
        if (elapsed_time >= BYTE_AVERAGE_REFRESH_SECONDS) {
            read_byte_sum_ /= 2;
            read_start_time_ = time(nullptr) - elapsed_time / 2;
        }
    }

    void Session::UpdateWriteByteAverage()
    {
        unsigned long elapsed_time = time(nullptr) - write_start_time_;
        if (elapsed_time >= BYTE_AVERAGE_REFRESH_SECONDS) {
            write_byte_sum_ /= 2;
            write_start_time_ = time(nullptr) - elapsed_time / 2;
        }
    }

    void Session::EnableEncryption()
    {
        encryption_ = true;
    }

    Encrypter& Session::encrypter()
    {
        return encrypter_;
    }

    tcp::socket& Session::socket()
    {
        return socket_;
    }

    UserID Session::id() const
    {
        return id_;
    }

    bool Session::login() const
    {
        return login_;
    }

    bool Session::online() const
    {
        return online_;
    }

    std::string Session::global_ip() const
    {
        return global_ip_;
    }

    uint16_t Session::udp_port() const{
        return udp_port_;
    }

    void Session::set_global_ip(const std::string& global_ip)
    {
        global_ip_ = global_ip;
    }

    void Session::set_udp_port(uint16_t udp_port)
    {
        udp_port_ = udp_port;
    }

    int Session::account_revision() const
    {
        return account_revision_;
    }

    void Session::set_account_revision(int revision)
    {
        account_revision_ = revision;
    }

    int Session::serialized_byte_sum() const
    {
        return serialized_byte_sum_;
    }

    int Session::compressed_byte_sum() const
    {
        return compressed_byte_sum_;
    }

    bool Session::operator==(const Session& s)
    {
        return id_ == s.id_;
    }

    bool Session::operator!=(const Session& s)
    {
        return !operator==(s);
    }

    std::string Session::Serialize(const Command& command)
    {
        unsigned int header = command.header();
        std::string body = command.body();

        serialized_byte_sum_ += body.size();

        // 圧縮
        if (!static_cast<bool>(header & COMPRESSED_FLAG) &&
                body.size() >= COMPRESS_MIN_LENGTH) {
            body = Utils::SnappyCompress(body);
            header |= COMPRESSED_FLAG;
        }

        compressed_byte_sum_ += body.size();

        std::string msg(reinterpret_cast<char*>(&header), sizeof(int));
        msg += body;

        // 暗号化
        if (encryption_) {
            msg = encrypter_.Encrypt(msg);
        }

        return Utils::Encode(msg);
    }

    Command Session::Deserialize(const std::string& msg)
    {
        std::string decoded_msg = Utils::Decode(msg);

        // 復号化
        if (encryption_) {
            decoded_msg = encrypter_.Decrypt(decoded_msg);
        }

        char header_buf[sizeof(int)];
        decoded_msg.copy(header_buf, sizeof(int));
        header::CommandHeader header = *reinterpret_cast<header::CommandHeader*>(header_buf);
        std::string body = decoded_msg;
        body.erase(0, sizeof(int));

        // 伸長
        if (static_cast<bool>(header & COMPRESSED_FLAG)) {
            body = Utils::SnappyUncompress(body);
            header = static_cast<header::CommandHeader>(header & ~(COMPRESSED_FLAG));
        }

        return Command(header, body, shared_from_this());
    }

    void Session::Receive(const boost::system::error_code& error)
    {
        if (!error) {
            std::string buffer(boost::asio::buffer_cast<const char*>(receive_buf_.data()),receive_buf_.size());
            auto length = buffer.find_last_of(Utils::DELIMITOR);

            if (length != std::string::npos) {

                receive_buf_.consume(length+1);
                buffer.erase(length+1);

                while (!buffer.empty()) {
                    std::string msg;

                    while (!buffer.empty() && buffer[0]!=Utils::DELIMITOR)
                    {
                        msg += buffer[0];
                        buffer.erase(0,1);
                    }
                    buffer.erase(0,1);

                    read_byte_sum_ += msg.size();
                    UpdateReadByteAverage();

                    Fetch(msg);
                }

                boost::asio::async_read_until(socket_,
                    receive_buf_, Utils::DELIMITOR,
                    boost::bind(
                      &Session::Receive, shared_from_this(),
                      boost::asio::placeholders::error));

            }

        } else {
            FatalError();
        }
    }

    void Session::DoWrite(const std::string msg)
    {
        bool write_in_progress = !send_queue_.empty();
        send_queue_.push(msg);
        if (!write_in_progress)
        {
          boost::asio::async_write(socket_,
              boost::asio::buffer(send_queue_.front().data(),
                send_queue_.front().size()),
              boost::bind(&Session::Write, this,
                boost::asio::placeholders::error));
        }
    }

    void Session::Write(const boost::system::error_code& error)
    {
        if (!error) {
            if (!send_queue_.empty()) {
                  send_queue_.pop();
                  if (!send_queue_.empty())
                  {
                    boost::asio::async_write(socket_,
                        boost::asio::buffer(send_queue_.front().data(),
                          send_queue_.front().size()),
                        boost::bind(&Session::Write, this,
                          boost::asio::placeholders::error));
                  }
            }
        } else {
            FatalError();
        }
    }

    void Session::Fetch(const std::string& msg)
    {
        if (msg.size() >= sizeof(int)) {
            if (on_receive_) {
                (*on_receive_)(Deserialize(msg));
            }
        }
    }

    void Session::FatalError()
    {
        online_ = false;
        if (on_receive_) {
            if (id_ > 0) {
                (*on_receive_)(FatalConnectionError(id_));
            } else {
                (*on_receive_)(FatalConnectionError());
            }
        }
    }

    void Session::set_on_receive(CallbackFuncPtr func)
    {
        on_receive_ = func;
    }


}
